library nueli.core.event;

import 'dart:async';

import 'package:flutter/cupertino.dart';

import 'data.dart' show $, Ref, OR, Add;
import 'util/--util.dart' show ORList;

// Events
typedef Construct<T> = T Function(); // Interlaced with setters
typedef Init<T> = dynamic Function(T data);
typedef Modify<T> = T Function(T data);
typedef Use<T> = void Function(T data); // Interlaced with getters
typedef Assign<K, T> = void Function(K key, T data);
typedef Query<K, T> = T Function(K key, [T fallback]);
typedef Convert<D, T> = D Function(T data);
typedef Test<T> = bool Function(T data);

// Async Events
typedef AsyncConstruct<T> = Future<T> Function();
typedef AsyncInit<T> = Future<dynamic> Function(T data);
typedef AsyncModify<T> = Future<T> Function(T data);
typedef AsyncUse<T> = Future<void> Function(T data);
typedef AsyncAssign<K, T> = Future<void> Function(K key, T data);
typedef AsyncQuery<K, T> = Future<T> Function(K key, [T fallback]);
typedef AsyncConvert<D, T> = Future<D> Function(T data);
typedef AsyncTest<T> = Future<bool> Function(T data);

/// Event exception
extension EventException on Exception {
  get msg => this.toString().replaceFirst('Exception: ', '');
}

/// Event Exception Handler
T handle<T>(
  T Function() event, {
  T Function(String msg) alt,
  T Function(T condition, bool success) ult,
}) {
  T _alt(String e) => (alt ?? (e) => null)(e);
  T _ult(T c, bool s) => (ult ?? (c, s) => c)(c, s);
  try {
    return _ult(event(), true);
  } catch (e) {
    try {
      return _ult(_alt((e as Exception).msg), false);
    } catch (e) {
      return _ult(_alt(''), false);
    }
  }
}

/// Async Event Exception Handler
Future<T> asyncHandle<T>(
  FutureOr<T> Function() task, {
  T Function(String msg) alt,
  FutureOr<T> Function(FutureOr<T> condition, bool success) ult,
}) async {
  T _alt(String e) => (alt ?? (e) => null)(e);
  FutureOr<T> _ult(FutureOr<T> c, bool s) => (ult ?? (c, s) async => await c)(c, s);
  try {
    return await _ult(await task(), true);
  } catch (e) {
    try {
      return _ult(_alt((e as Exception).msg), false);
    } catch (e) {
      return _ult(_alt(''), false);
    }
  }
}

/// Event Cycle of a Typical Data
class Cycle<T> with OR<Cycle<T>> {
  final Construct<T> construct;
  final Init<T> init;
  final Modify<T> modify;
  final Use<T> use;
  final Function() destroy;

  const Cycle({
    this.construct,
    this.init,
    this.modify,
    this.use,
    this.destroy,
  });

  @override
  Cycle<T> operator |(Cycle<T> element) {
    return Cycle(
      construct: () {
        final obj = (construct ?? () => null)();
        return (element.construct ?? () => null)() ?? obj;
      },
      init: (data) {
        final result = (init ?? (data) => data)(data);
        final result2 = (element.init ?? (data) => data)(data is T ? result : data);
        return data is T ? result2 : data;
      },
      modify: (data) {
        final result = (modify ?? (data) => data)(data);
        return (element.modify ?? (data) => data)(result);
      },
      use: (data) {
        (use ?? (data) {})(data);
        (element.use ?? (data) {})(data);
      },
      destroy: () {
        (destroy ?? () {})();
        (element.destroy ?? () {})();
      },
    );
  }
}

/// Async Event Cycle of a Typical Data
class AsyncCycle<T> with OR<AsyncCycle<T>> {
  final AsyncConstruct<T> construct;
  final AsyncInit<T> init;
  final AsyncModify<T> modify;
  final AsyncUse<T> use;
  final Future Function() destroy;

  const AsyncCycle({
    this.construct,
    this.init,
    this.modify,
    this.use,
    this.destroy,
  });

  @override
  AsyncCycle<T> operator |(AsyncCycle<T> element) {
    return AsyncCycle(
      construct: () async {
        final obj = await (construct ?? () async => null)();
        return await (element.construct ?? () async => null)() ?? obj;
      },
      init: (data) async {
        final result = await (init ?? (data) async => data)(data);
        final result2 = await (element.init ?? (data) async => data)(data is T ? result : data);
        return data is T ? result2 : data;
      },
      modify: (data) async {
        final result = await (modify ?? (data) async => data)(data);
        return await (element.modify ?? (data) async => data)(result);
      },
      use: (data) async {
        await (use ?? (data) async {})(data);
        await (element.use ?? (data) async {})(data);
      },
      destroy: () async {
        await (destroy ?? () async {})();
        await (element.destroy ?? () async {})();
      },
    );
  }
}

/// Typical Data with Event cycle implementation
mixin CycleFeed<T> {
  final Ref<List<Cycle<T>>> _cycle = $([]);

  Cycle<T> get cycle {
    return handle(
      () => [
        ..._cycle.data,
        Cycle(construct: construct, init: init, modify: modify, use: use, destroy: destroy),
      ].foldThrough(),
      alt: (e) => Cycle(),
    );
  }

  set cycle(Cycle<T> cycle) => _cycle.data.add(cycle);

  T construct();

  @mustCallSuper
  init(T data) {}

  T modify(T data) => data;

  @mustCallSuper
  void use(T data) {}

  @mustCallSuper
  void destroy() {}
}

/// Event Cycle Mixin
mixin CycleReel<T> {
  T get data;
  set data(T data);
  T get fallback => null;
  Cycle<T> get cycle;

  @mustCallSuper
  void construct([Construct<T> construct]) {
    if (cycle.construct != null) data = cycle.construct();
    if (construct != null) data = construct();
  }

  @mustCallSuper
  void init([Init<T> config]) {
    if (cycle.init != null) cycle.init(data ?? fallback);
    if (config != null) config(data ?? fallback);
  }

  @mustCallSuper
  void modify([Modify<T> modify]) {
    if (cycle.modify != null) data = cycle.modify(data ?? fallback);
    if (modify != null) data = modify(data ?? fallback);
  }

  @mustCallSuper
  void use([Use<T> use]) {
    if (cycle.use != null) cycle.use(data ?? fallback);
    if (use != null) use(data ?? fallback);
  }

  void destroy() {
    if (cycle.destroy != null) cycle.destroy();
    data = fallback;
  }
}

/// Async Event Cycle Mixin
mixin AsyncCycleReel<T> {
  T get data;
  set data(T data);
  T get fallback => null;
  AsyncCycle<T> get cycle;

  @mustCallSuper
  Future<void> construct([AsyncConstruct<T> construct]) async {
    if (cycle.construct != null) data = await cycle.construct();
    if (construct != null) data = await construct();
  }

  @mustCallSuper
  Future<void> init([AsyncInit<T> config]) async {
    var obj;
    if (cycle.init != null) obj = await cycle.init(data ?? fallback);
    if (obj is T) this.data = obj;
    if (config != null) obj = await config(data ?? fallback);
    if (obj is T) this.data = obj;
  }

  @mustCallSuper
  Future<void> modify([AsyncModify<T> modify]) async {
    if (cycle.modify != null) data = await cycle.modify(data ?? fallback);
    if (modify != null) data = await modify(data ?? fallback);
  }

  @mustCallSuper
  Future<void> use([AsyncUse<T> use]) async {
    if (cycle.use != null) await cycle.use(data ?? fallback);
    if (use != null) await use(data ?? fallback);
  }

  @mustCallSuper
  void destroy() {
    if (cycle.destroy != null) cycle.destroy();
    data = fallback;
  }
}

/// Shortcut to begin an event Algorithm
Algorithm get algog => Algorithm();

/// Event Algorithm <br>
/// <span color = '#99aaff'> Begin an algorithm</span> | [ag & async {task}] <br>
/// <span color = '#99aaff'> Create an empty algorithm<span> | [ag & null ] <br>
class Algorithm<T> with Add<Algorithm<T>> {
  final Ref<List<Future<T> Function()>> _flow = $([]);

  Algorithm([List<Future<T> Function()> flow = const []]) {
    _flow.data = flow.toList();
  }

  Algorithm<T> operator &(Future<T> Function() task) {
    if (task == null) {
      return Algorithm(_flow.data);
    } else
      return Algorithm<T>([..._flow.data, task]);
  }

  @override
  Algorithm<T> operator +(Algorithm<T> algorithm) {
    if (algorithm == null) {
      return Algorithm(_flow.data);
    } else {
      return Algorithm([..._flow.data, ...algorithm._flow.data]);
    }
  }

  Future<T> call({bool reversed}) async {
    T result;
    for (Future<T> Function() task in _flow.data) {
      await task().then((value) => result = value);
    }
    return result;
  }
}
