library nueli.core.util;

import 'dart:io';
import 'dart:math';

import '../data.dart' show Add, Subtract, Multiply, Divide, OR, AND;

part 'file.dart';

double modulus(num value) => sqrt(value * value);
double slope(Point a, Point b) => a.distanceTo(b);

extension AdvancedDateTime on DateTime {
  static const daysOfTheWeek = ['Monday', 'Tuesday', 'Wednessday', 'Thursday', 'Friday', 'Saturday', 'Sunday'];
  static const monthsOfTheYear = ['January', 'Febuary', 'March', 'April', 'May', 'June', 'July', 'August', 'September', 'October', 'November', 'December'];
  static const daysOfTheWk = ['Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat', 'Sun'];
  static const monthsOfTheYr = ['Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun', 'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec'];

  /// parse the specified [date] according to the specified [format]. <br>
  /// eg. 'DD/MM/YYY  hh/mm a' <br><br>
  /// <style color='ffa836'>[Day] : D\>1, DD\>01, DDD\>Mon, DDDD\>Monday <br><br>
  /// <style color='ffa836'>[Month] : M\>[1], MM\>01, MMM\>Jan, MMMM\>January <br><br>
  /// <style color='ffa836'>[Year] : Y\>21, YY\>21, YYY\>2021, YYYY\>2021 <br><br>
  /// <style color='ffa836'>[A] : aa\>am, A\>AM, aa\>am, AA\>AM, Aa\>Am, aA\>aM <br><br>
  String parse([String format]) {
    return '$day ${monthsOfTheYr[month - 1]}, $year ${hour % 12} : $minute ${hour > 12 ? 'PM' : 'AM'}';
  }
}

extension AdvanceList<T> on List<T> {
  bool get isUnitary => length == 1;
  bool isMemberOf(List<T> list, {int offset = 0, int breakOffset}) => list.includes(this);
  bool startsWith(List<T> list, [int offset = 0]) => includes(list, offset: offset, breakOffset: offset);
  bool endsWith(List<T> list, [int offset = 0]) => reversed.toList().includes(list.reversed, offset: offset, breakOffset: offset);

  bool includes(List<T> list, {int offset = 0, int breakOffset}) {
    if (list.isEmpty) return true;
    if (list.isUnitary) return contains(list[0]);
    int notch = offset ?? 0;
    if (length < list.length) return false;
    for (int i = 0; i < length; i++) {
      T item = this[i];
      if (notch >= list.length) return true;
      if (item != list[notch]) {
        if (breakOffset != null && notch >= breakOffset) return false;
        notch = 0;
        if (length - i - 1 >= list.length) continue;
        return false;
      }
      notch++;
    }
    return false;
  }

  List<T> serialize<D extends Comparable>([D Function(T) converter]) {
    final _converter = converter ?? (t) => null;
    List<T> result = [];
    List<D> serialList = [];
    for (T item in this) {
      if (!serialList.contains(_converter(item))) {
        result.add(item);
        serialList.add(_converter(item));
      }
    }
    return result;
  }

  List<T> truncate(T item, [List<T> others]) {
    List<T> list = [];
    forEach((element) => list..add(element)..add(item));
    return list..removeLast();
  }

  List<T> rebind([bool Function(T) validator]) {
    List<T> list = [];
    forEach((item) => (validator ?? (i) => i != null)(item) ? list.add(item) : {});
    return list;
  }
}

extension ListOfList<T> on List<List<T>> {
  List<T> spread() {
    List<T> result = [];
    for (List<T> list in this) result.addAll(list);
    return result;
  }
}

// operations
extension SubtractList<T extends Subtract<T>> on List<T> {
  List<T> minus(List<T> operand) {
    List<T> transforms = [];
    for (int i = 0; i < max(length, operand.length); i++) {
      if (i < length && i < operand.length)
        transforms.add(this[i] - operand[i]);
      else if (i < length)
        transforms.add(this[i]);
      else if (i < operand.length) transforms.add(operand[i]);
    }
    return transforms;
  }
}

extension AddList<T extends Add<T>> on List<T> {
  List<T> plus(List<T> operand) {
    List<T> transforms = [];
    for (int i = 0; i < max(length, operand.length); i++) {
      if (i < length && i < operand.length)
        transforms.add(this[i] + operand[i]);
      else if (i < length)
        transforms.add(this[i]);
      else if (i < operand.length) transforms.add(operand[i]);
    }
    return transforms;
  }

  T addThrough() {
    if (length < 1) throw ('cannot add through an empty list');
    return (this..removeAt(0)).fold(this[0], (previousValue, element) => previousValue + element);
  }
}

extension MultiplyList<T extends Multiply<T>> on List<T> {
  List<T> multiplyBy(List<T> operand) {
    List<T> transforms = [];
    for (int i = 0; i < max(length, operand.length); i++) {
      if (i < length && i < operand.length)
        transforms.add(this[i] * operand[i]);
      else if (i < length)
        transforms.add(this[i]);
      else if (i < operand.length) transforms.add(operand[i]);
    }
    return transforms;
  }

  T multiplyThrough() {
    if (length < 1) throw ('cannot multiply through an empty list');
    return (this..removeAt(0)).fold(this[0], (previousValue, element) => previousValue * element);
  }
}

extension DivideList<T extends Divide<T>> on List<T> {
  List<T> divideBy(List<T> operand) {
    List<T> transforms = [];
    for (int i = 0; i < max(length, operand.length); i++) {
      if (i < length && i < operand.length)
        transforms.add(this[i] / operand[i]);
      else if (i < length)
        transforms.add(this[i]);
      else if (i < operand.length) transforms.add(operand[i]);
    }
    return transforms;
  }

  T divideThrough() {
    if (length < 1) throw ('cannot divide through an empty list');
    return (this..removeAt(0)).fold(this[0], (previousValue, element) => previousValue / element);
  }
}

extension ORList<T extends OR<T>> on List<T> {
  List<T> or(List<T> operand) {
    List<T> transforms = [];
    for (int i = 0; i < max(length, operand.length); i++) {
      if (i < length && i < operand.length)
        transforms.add(this[i] | operand[i]);
      else if (i < length)
        transforms.add(this[i]);
      else if (i < operand.length) transforms.add(operand[i]);
    }
    return transforms;
  }

  T foldThrough() {
    if (length < 1) throw ('cannot fold through an empty list');
    return (this..removeAt(0)).fold(this[0], (previousValue, element) => previousValue | element);
  }
}

extension ANDList<T extends AND<T>> on List<T> {
  List<T> and(List<T> operand) {
    List<T> transforms = [];
    for (int i = 0; i < max(length, operand.length); i++) {
      if (i < length && i < operand.length)
        transforms.add(this[i] & operand[i]);
      else if (i < length)
        transforms.add(this[i]);
      else if (i < operand.length) transforms.add(operand[i]);
    }
    return transforms;
  }

  T concatThrough() {
    if (length < 1) throw ('cannot concat through an empty list');
    return (this..removeAt(0)).fold(this[0], (previousValue, element) => previousValue & element);
  }
}
